/****************************************
 *      the Beos Reversing Engine       * 
 *              by Zadig                *
 *                                      *
 * This file contains asm parsing       *
 * functions.                           *
 ****************************************/

/*these files must always be includes before all others*/
#include <types.h>
#include "globals.h"
#include "structs.h"
#include "reveng_errors.h"
/* end these files must always be includes before all others*/

#include <zem_errors.h>
#include <stdio.h>
#include <string.h>

#include "reveng.h"
#include "objdump.h"
#include "asm.h"
#include "misc.h"

/***********************************************************
 *  Function: ASM_CreateAsmFile               File: asm.c  *
 ***********************************************************
 * Usage:
 *
 * Creates a file containing only the asm instructions.
 ***********************************************************/ 
int32 ASM_CreateAsmFile(Settings *projectSettings)
{
   FILE *f_DisasmFile,*f_AsmFile;
   uint8 sz_CurLine[STRING_MAX_SIZE];
   uint8 sz_LineToWrite[STRING_MAX_SIZE];
   int32 i_ret;
   uint8 b_FirstInst=Z_FALSE;

   uint8 sz_Addr[STRING_MAX_SIZE],sz_HexCode[STRING_MAX_SIZE],sz_Instruction[STRING_MAX_SIZE];
   uint8 sz_PrevAddr[STRING_MAX_SIZE],sz_PrevHexCode[STRING_MAX_SIZE],sz_PrevInstruction[STRING_MAX_SIZE];

   DBG_PRINT("Entering RE_CreateAsmFile...\n");  
   f_DisasmFile = fopen((char*)(projectSettings->objdump_FileName),"r"); 
   if(f_DisasmFile == NULL)
   {
      return(FILE_ERROR_OPEN);
   }

   f_AsmFile =  fopen((char*)(projectSettings->asm_SrcFileName),"w+"); 
   if(f_AsmFile == NULL)
   {
      return(FILE_ERROR_OPEN);
   }

   /* goto begining of section */
   do
   {
      i_ret = RE_ReadString(sz_CurLine,STRING_MAX_SIZE,f_DisasmFile);
      if(i_ret != Z_OK)
      {
         return(FILE_ERROR_READ);
      }
   } while(strncmp(sz_CurLine,"Disassembly of section",20) != 0) ;

   RE_ReadString(sz_CurLine,STRING_MAX_SIZE,f_DisasmFile);
   RE_ReadString(sz_CurLine,STRING_MAX_SIZE,f_DisasmFile);
   i_ret = RE_ReadString(sz_CurLine,STRING_MAX_SIZE,f_DisasmFile);

   while(i_ret == Z_OK)
   {
      /* if new function */
      if(sz_CurLine[0] == '\0')
      {
         RE_ReadString(sz_CurLine,STRING_MAX_SIZE,f_DisasmFile);
         i_ret = RE_ReadString(sz_CurLine,STRING_MAX_SIZE,f_DisasmFile);        
         if(i_ret != Z_OK)
         {
            return(FILE_ERROR_READ);
         }
      }
      else if(sz_CurLine[0] == 'D')
      {
         RE_ReadString(sz_CurLine,STRING_MAX_SIZE,f_DisasmFile);
         RE_ReadString(sz_CurLine,STRING_MAX_SIZE,f_DisasmFile);
         i_ret = RE_ReadString(sz_CurLine,STRING_MAX_SIZE,f_DisasmFile);
         if(i_ret != Z_OK)
         {
            return(FILE_ERROR_READ);
         }
      }

      RE_DecomposeInstruction(sz_CurLine, sz_Addr, sz_HexCode, sz_Instruction);
      if(sz_Instruction[0] == '\0')
      {
         strcat(sz_PrevHexCode,sz_HexCode);
      }

      if(b_FirstInst == Z_FALSE)
      {        
         b_FirstInst = Z_TRUE;
      }
      else
      {
         sprintf(sz_LineToWrite,"%s:  %20s     ",sz_PrevAddr, sz_PrevHexCode);      
         strcat(sz_LineToWrite,sz_PrevInstruction);
         strcat(sz_LineToWrite,"\n\0");

         if(sz_Instruction[0] != '\0')
         {
            fputs(sz_LineToWrite,f_AsmFile);
         }
      }

      if(sz_Instruction[0] != '\0')
      {
         strcpy(sz_PrevAddr,sz_Addr);
         strcpy(sz_PrevHexCode,sz_HexCode);
         strcpy(sz_PrevInstruction,sz_Instruction);                
      }      

      i_ret = RE_ReadString(sz_CurLine,STRING_MAX_SIZE,f_DisasmFile);      
   } /* end while i_ret */

   sprintf(sz_LineToWrite,"%s:  %20s     ",sz_Addr, sz_HexCode);      
   strcat(sz_LineToWrite,sz_Instruction);
   strcat(sz_LineToWrite,"\n\0");
   fputs(sz_LineToWrite,f_AsmFile);

   fclose(f_DisasmFile);
   fclose(f_AsmFile);
   DBG_PRINT("End of RE_CreateAsmFile\n");    
   return(Z_OK);
}

/***********************************************************
 *  Function: ASM_ProcessInstruction          File: asm.c  *
 ***********************************************************
 * Usage:
 *
 * Analyse an asm instruction and put the result in
 * i_Result. If the result is not valid then i_ResultType
 * is RESULT_TYPE_INVALID
 ***********************************************************/ 
int32 ASM_ProcessInstruction(Register *registers, 
      RegisterContext *context,
      uint8 *sz_Instruction, 
      Mapping  *ps_CurFunction,
      uint32 i_Addr, 
      uint32 *i_Result,
      RE_ResultType *i_ResultType,
      uint8 b_CheckForMessages,
      Function_Params *parameters)
{
   uint8 i_ParamIndex;
   uint8 b_End, b_StillInFirst;
   uint8 *psz_Instruction;
   uint8 sz_Opcode[STRING_MAX_SIZE];
   uint8 sz_Param1[STRING_MAX_SIZE];
   uint8 sz_Param2[STRING_MAX_SIZE];
   uint32 i_Param1;
   uint32 i_Index, i_CntxIndex;
   RE_Registers i_IndexReg;
   RE_ResultType i_Param1Type;
   
   /* initialisation */
   *i_Result = 0;
   *i_ResultType = RESULT_TYPE_INVALID;
   i_Param1Type = RESULT_TYPE_INVALID; 
   i_Index = 0;
   psz_Instruction = sz_Instruction;
   /* end initialisation */

   /* update registers if necessary */
   i_CntxIndex = context->nextCntx;
   if( (i_CntxIndex != RE_DISABLED) &&
       (i_Addr == context->context[i_CntxIndex].addr) )
   {
      memcpy(registers, &(context->context[i_CntxIndex].registers[0]), 
             REG_MAX * sizeof(Register) );
      RE_FreeCntx(context, i_CntxIndex);
      RE_SetNextCntx(context);
   }
   /* gets opcode */ 
   while( (*psz_Instruction != '\0') && (*psz_Instruction != ' ') )
   {
      sz_Opcode[i_Index] = *psz_Instruction;
      i_Index++;
      psz_Instruction++;
   }
   sz_Opcode[i_Index] = '\0';

   while(*psz_Instruction == ' ') 
      psz_Instruction++;

   /* instruction without parameters */
   if( (*psz_Instruction == '\0') || (*psz_Instruction == '\n') )
   {
      *i_ResultType = RESULT_TYPE_INVALID;
      return(Z_OK);
   }

   /* gets first param */
   i_Index = 0;
   b_End = Z_FALSE;
   b_StillInFirst = Z_FALSE;
   while( b_End == Z_FALSE )
   {
      sz_Param1[i_Index] = *psz_Instruction;
      i_Index++;
      psz_Instruction++;

      if(*psz_Instruction == '(')
         b_StillInFirst = Z_TRUE;
      else if(*psz_Instruction == ')')
         b_StillInFirst = Z_FALSE;
      else if( (b_StillInFirst == Z_FALSE) && (*psz_Instruction == ',') )
         b_End = Z_TRUE;
      else if(*psz_Instruction == '\n')
         b_End = Z_TRUE;
   }  
   sz_Param1[i_Index] = '\0';

   if(*psz_Instruction != '\0')
      psz_Instruction++;

   /* gets second param */
   i_Index = 0;
   while( (*psz_Instruction != '\0') && (*psz_Instruction != ',') )
   {
      sz_Param2[i_Index] = *psz_Instruction;
      i_Index++;
      psz_Instruction++;
   }  
   sz_Param2[i_Index] = '\0';

   ASM_GetValueOfParam(registers, sz_Param1, &i_Param1, &i_Param1Type);  
   *i_ResultType = i_Param1Type;

   if(STREQ3(sz_Opcode,"ret"))
      RE_InitRegisters(registers, Z_TRUE);

   if(i_Param1Type == RESULT_TYPE_INVALID)
   {
      /* if instruction is pop or push, update stack */
      if(STREQ3(sz_Opcode,"pop")) 
      {
         registers[REG_ESP].value += 4;
         if(STREQ4(sz_Param1,"%ebx"))
         {
            registers[REG_EBX].isValid = Z_TRUE;
            registers[REG_EBX].type = RESULT_TYPE_DIRECT_VALUE;
            registers[REG_EBX].value = i_Addr;
            *i_Result = i_Addr;
         }
      }
      else if(STREQ4(sz_Opcode,"push")) 
         registers[REG_ESP].value -= 4;   

      /* if dest is a register we must invalidate it */
      else if(sz_Param2[0] == '%')
      {
         i_IndexReg = ASM_GetRegisterIndex(registers, sz_Param2);
         if(i_IndexReg != REG_MAX)
         {
            registers[i_IndexReg].isValid = Z_FALSE;
            registers[i_IndexReg].type = RESULT_TYPE_INVALID;
         }       
      }   
      return(Z_OK);
   }

   /* now let's analyse the code */
   if( (b_CheckForMessages == Z_TRUE) &&
         (i_Param1Type == RESULT_TYPE_DIRECT_VALUE) &&
         (STREQ3(sz_Opcode,"cmp")) )
   {
      *i_Result = i_Param1;

      if(sz_Param2[0] == '(')
         i_IndexReg = ASM_GetRegisterIndex(registers, sz_Param2+1);  /* cmp xxx,(%reg) */
      else
         i_IndexReg = ASM_GetRegisterIndex(registers, sz_Param2);  /* cmp xxx,%reg */

      if( (i_IndexReg != REG_MAX) &&
            (registers[i_IndexReg].type == RESULT_TYPE_PARAM) )
      { 
         i_ParamIndex = RE_GetParamIndex(parameters, registers[i_IndexReg].value);
         if( (i_ParamIndex != 0) &&
               (RE_ParamIsMessage(parameters, i_ParamIndex) == Z_TRUE) )
         {
            *i_ResultType = RESULT_TYPE_MESSAGE;
         }
      }
   }
   else if( (STREQ3(sz_Opcode,"mov")) ||
         (STREQ3(sz_Opcode,"lea")) ) /* not correct but does what I want for the moment */
   { 
      if(i_Param1Type == RESULT_TYPE_PARAM)
         *i_Result = i_Param1;

      /* look for register index */
      i_IndexReg = ASM_GetRegisterIndex(registers, sz_Param2);
      if(i_IndexReg != REG_MAX)
      {
         registers[i_IndexReg].value = i_Param1;
         registers[i_IndexReg].isValid = Z_TRUE;
         registers[i_IndexReg].type = i_Param1Type;
         *i_Result = i_Param1;
      }     
   }
   else if(STREQ3(sz_Opcode,"add")) 
   {     
      i_IndexReg = ASM_GetRegisterIndex(registers, sz_Param2);
      if( (i_IndexReg != REG_MAX) &&
            (registers[i_IndexReg].isValid == Z_TRUE) )
      {
         registers[i_IndexReg].value += i_Param1;
      }
   }
   else if(STREQ3(sz_Opcode,"sub")) 
   {
      i_IndexReg = ASM_GetRegisterIndex(registers, sz_Param2);
      if( (i_IndexReg != REG_MAX) &&
            (registers[i_IndexReg].isValid == Z_TRUE) )
      {
         registers[i_IndexReg].value -= i_Param1;
      }
   }
   else if(STREQ4(sz_Opcode,"call"))
   {      
      *i_Result = i_Param1;
      RE_InitRegisters(registers, Z_FALSE);
      /* some versions of gcc do the pop ebx in a call to a function containing
         only "mov    (%esp,1),%ebx" and a "ret". This function seem to be
         always just before the function start, ie 4 bytes before function start
         */
      if( (ps_CurFunction != NULL) &&
          (i_Param1 == ps_CurFunction->addr - 4) )
      {
         registers[REG_EBX].isValid = Z_TRUE;
         registers[REG_EBX].type = RESULT_TYPE_DIRECT_VALUE;
         registers[REG_EBX].value = i_Addr + 5; /* current addr + call opcode size */
         *i_Result = registers[REG_EBX].value;
      }
   }
   else if(sz_Opcode[0] == 'j')
   {
      *i_Result = i_Param1;
      *i_ResultType = RESULT_TYPE_JUMP;
      if(i_Param1 > i_Addr)
      {
         RE_AddCntx(context, registers, i_Param1);
      }

      if(STREQ3(sz_Opcode,"jmp"))
      {
         RE_InitRegisters(registers, Z_FALSE);
      }
   }
   else if(STREQ3(sz_Opcode,"pop")) 
   {
      registers[REG_ESP].value += 4;
      if(strncmp(sz_Param1,"%ebx",4) == 0)
      {
         registers[REG_EBX].isValid = Z_TRUE;
         registers[REG_EBX].value = i_Addr;
         registers[REG_EBX].type = RESULT_TYPE_DIRECT_VALUE;
         *i_Result = i_Addr;
      }
      else
         *i_ResultType = RESULT_TYPE_INVALID;
   }
   else if(STREQ4(sz_Opcode,"push")) 
   {
      registers[REG_ESP].value -= 4;

      i_IndexReg = ASM_GetRegisterIndex(registers, sz_Param1);
      if( (i_IndexReg != REG_MAX) &&
            (registers[i_IndexReg].isValid == Z_TRUE) )
      {
         *i_Result = registers[i_IndexReg].value;
      }
      else if(sz_Param1[0] == '$')
         *i_Result = i_Param1;
   } 
   else
      *i_ResultType = RESULT_TYPE_INVALID;

   return(Z_OK);
}

/***********************************************************
 *  Function: ASM_GetValueOfParam             File: asm.c  *
 ***********************************************************
 * Usage:
 *
 * Gets the value of the parameter and checks if it is
 * valid or not.
 ***********************************************************/ 
int32 ASM_GetValueOfParam(Register *registers, uint8 *sz_Param, 
                          uint32 *i_Param, RE_ResultType *i_ParamType)
{
   RE_Registers i_IndexReg;
   uint32 i_Index;

   /* gets value of param 1 */
   if(sz_Param[0] == '\0')
   {
      *i_ParamType = RESULT_TYPE_INVALID;
   }
   /* immediate address */
   else if(sz_Param[0] == '$')
   {
      sscanf(sz_Param+3,"%X",(unsigned int*)i_Param);
      *i_ParamType = RESULT_TYPE_DIRECT_VALUE;
   }
   /* register */
   else if(sz_Param[0] == '%')
   {

      i_IndexReg = ASM_GetRegisterIndex(registers, sz_Param);
      if( (i_IndexReg != REG_MAX) &&
            (registers[i_IndexReg].isValid == Z_TRUE) )
      {
         *i_Param = registers[i_IndexReg].value;
         *i_ParamType = registers[i_IndexReg].type; /* RESULT_TYPE_INDIRECT_VALUE; */
      }        
   }
   /* offset on a register */
   else if( (sz_Param[0] == '0') && (sz_Param[1] == 'x') )
   {
      if(strstr(sz_Param,"(") != NULL)
      {
         sscanf(sz_Param+2,"%X",(unsigned int*)i_Param);
         i_Index = 2;
         while( (sz_Param[i_Index] != '(')  && 
               (sz_Param[i_Index] != '\0') && 
               (sz_Param[i_Index] != ',') )
         {
            i_Index++;
         }

         if(sz_Param[i_Index] == '(')
         {
            i_IndexReg = ASM_GetRegisterIndex(registers, &(sz_Param[i_Index+1]));
            if( ( (i_IndexReg == REG_EBP) || (i_IndexReg == REG_ESP) )&&   
                  ( (*i_Param & 0x80000000) == 0 ) )    
                  /* check that offset is positive          */
                  /* (EBP - xxx refers to local variables)  */
            {
               (*i_Param) += registers[i_IndexReg].value;
               *i_ParamType = RESULT_TYPE_PARAM;
            }
            else if( (i_IndexReg != REG_MAX) &&
                  (registers[i_IndexReg].isValid == Z_TRUE) )
            {
               (*i_Param) += registers[i_IndexReg].value;
               *i_ParamType = RESULT_TYPE_INDIRECT_VALUE;
            }        
         }   
      }
      else
      {
         sscanf(sz_Param+2,"%X",(unsigned int *)i_Param);
         *i_ParamType = RESULT_TYPE_DIRECT_VALUE;
      }
   }
   /* pointer value */
   /* We treat it only if it is a parameter (Usefull for messages) */
   else if(sz_Param[0] == '(')
   {
      i_IndexReg = ASM_GetRegisterIndex(registers, &(sz_Param[1]));
      if( (i_IndexReg < REG_MAX) &&
          (registers[i_IndexReg].type == RESULT_TYPE_PARAM) )
      {
         (*i_Param) = registers[i_IndexReg].value;
         *i_ParamType = RESULT_TYPE_PARAM;
      }        
   }   
   /* direct hexa value */
   else if( (sz_Param[0] != '(') &&
         (sz_Param[0] != '*') )
   {
      sscanf(sz_Param,"%X",(unsigned int *)i_Param);
      *i_ParamType = RESULT_TYPE_DIRECT_VALUE;
   }
   else
      *i_ParamType = RESULT_TYPE_INVALID;

   return(Z_OK);
}

/***********************************************************
 *  Function: ASM_GetRegisterIndex            File: asm.c  *
 ***********************************************************
 * Usage:
 *
 * Return the index of the register or REG_MAX if the
 * param is not a register.
 ***********************************************************/ 
RE_Registers ASM_GetRegisterIndex(Register *registers, uint8 *sz_Register)
{
   uint8 i;
   RE_Registers i_IndexReg;

   i_IndexReg = REG_MAX;

   if(sz_Register[0] == '%')  
   {
      i=0;
      while(i < REG_MAX)
      {
         if(strncmp(sz_Register+1, registers[i].name,3) == 0) 
         {
            i_IndexReg = i;
            break;
         }
         i++;
      }
   }

   return(i_IndexReg);
}
