/**********************************************
 *          the Beos Reversing Engine         * 
 *                  by Zadig                  *
 *                                            *
 * This file contains the chain-lists         *
 * manipulation functions.                    *
 **********************************************/

/*these files must always be includes before all others*/
#include <types.h>
#include "globals.h"
#include "structs.h"
#include "reveng_errors.h"
/* end these files must always be includes before all others */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "chainlist.h"
#include "misc.h"

/**************************************************************
 *                     Mapping functions                      *
 **************************************************************/ 

/***********************************************************
 *  Function: LIST_AlphaSort            File: chainlist.c  *
 ***********************************************************
 * Usage:
 *
 * Sorts all elements in alphabetical order
 ***********************************************************/ 
int32 LIST_AlphaSort(LIST_Mapping *list)
{
   uint8    b_Sorted;
   Mapping *ps_OldMap, *ps_CurMap, *ps_TmpMap;
   
   if(list->list_Map == NULL)
      return(CHAIN_ERROR_NOT_FOUND);
   
   ps_OldMap = list->list_Map;
   ps_CurMap = list->list_Map;
   
   ps_OldMap = ps_OldMap->next;

   ps_CurMap->next_alpha = NULL;
   ps_CurMap->prev_alpha = NULL;

   while(ps_OldMap != NULL)
   {        
      b_Sorted = Z_FALSE;
      while(b_Sorted == Z_FALSE)
      {
         /* insert */
         if(strcasecmp(ps_OldMap->label, ps_CurMap->label) < 0)
         {
            /* insert at the begining of the list */
            if(ps_CurMap->prev_alpha == NULL)
            {
               ps_TmpMap = ps_OldMap;
               ps_OldMap = ps_OldMap->next;
               
               ps_CurMap->prev_alpha = ps_TmpMap;
               ps_TmpMap->prev_alpha = NULL;
               ps_TmpMap->next_alpha = ps_CurMap;
               list->alpha = ps_TmpMap;
               b_Sorted = Z_TRUE;
            }
            /* new element is before */
            else if(strcasecmp(ps_OldMap->label, ps_CurMap->prev_alpha->label) < 0)
            {
               ps_CurMap = ps_CurMap->prev_alpha;
            }
            /* new element is just before the current */
            else
            {
               ps_TmpMap = ps_OldMap;
               ps_OldMap = ps_OldMap->next;
               
               ps_TmpMap->prev_alpha = ps_CurMap->prev_alpha;
               ps_TmpMap->next_alpha = ps_CurMap;
               ps_CurMap->prev_alpha = ps_TmpMap;
               ps_TmpMap->prev_alpha->next_alpha = ps_TmpMap;
               b_Sorted = Z_TRUE;
            }
         }
         /* new element is after the current one */
         else
         {
            /* add at the end of the list */
            if(ps_CurMap->next_alpha == NULL)
            {
               ps_TmpMap = ps_OldMap;
               ps_OldMap = ps_OldMap->next;
               
               ps_CurMap->next_alpha = ps_TmpMap;
               ps_TmpMap->prev_alpha = ps_CurMap;
               ps_TmpMap->next_alpha = NULL;
               b_Sorted = Z_TRUE;
            }
            else
               ps_CurMap = ps_CurMap->next_alpha;
         }
      }
   }
 
   return(Z_OK);
}

/***********************************************************
 *  Function: LIST_FreeMap              File: chainlist.c  *
 ***********************************************************
 * Usage:
 *
 * Frees all elements of a list.
 ***********************************************************/ 
int32 LIST_FreeMap(LIST_Mapping *list)
{
   Mapping *p_Next,*p_Map = list->list_Map;

   if( p_Map == NULL )
   {
      return(Z_OK);
   }

   while (p_Map->next != NULL)
   {
      p_Next = p_Map->next;
      LIST_FreeRef(&(p_Map->references));
      LIST_FreeCallees(&(p_Map->callees));
      free(p_Map);
      p_Map=p_Next;

   }

   free(p_Map);
   list->list_Map = NULL;
   list->alpha = NULL;

   return(Z_OK);
}

/***********************************************************
 *  Function: LIST_DeleteMap            File: chainlist.c  *
 ***********************************************************
 * Usage:
 *
 * Removes an element of the list.
 ***********************************************************/ 
int32 LIST_DeleteMap(LIST_Mapping *list, Mapping *mapping)
{
   mapping->prev->next = mapping->next;
   mapping->next->prev = mapping->prev;

   if(list->list_Map == mapping)    
      list->list_Map = mapping->next;

   if(list->tail == mapping)
      list->tail = mapping->prev;

   free(mapping);
   list->alpha = NULL;

   return(Z_OK);
}

/***********************************************************
 *  Function: LIST_AddMap            File: chainlist.c  *
 ***********************************************************
 * Usage:
 *
 * Adds an element in the list and sort it by address.
 ***********************************************************/ 
Mapping *LIST_AddMap(LIST_Mapping *list, Mapping map)
{
   Mapping *p_Mess, *p_Prev, *p_Cur;

   p_Prev = NULL;
   /* Allocates the object */
   p_Mess = (Mapping*)calloc(1, sizeof(Mapping));

   if (list->list_Map == NULL)
   {
      list->list_Map = p_Mess; /*(Mapping*) malloc(sizeof(Mapping));*/
      list->list_Map->addr = map.addr;
      list->list_Map->size = map.size;
      list->list_Map->is_printed = Z_FALSE;
      strcpy(list->list_Map->label, map.label);
      list->list_Map->type = map.type;
      memcpy(&(list->list_Map->references),&(map.references),sizeof(LIST_Reference));
      list->list_Map->next = NULL;
      list->list_Map->prev = NULL;
      list->list_Map->references.list_Ref = NULL;
      list->tail = list->list_Map;
      return(p_Mess);
   }

   p_Cur = list->tail;
   if (map.addr >= p_Cur->addr)
   {
      p_Mess->next = NULL;
      p_Mess->prev = p_Cur;
      p_Cur->next = p_Mess;
      list->tail = p_Mess;
   }
   else
   {
      while( (map.addr < p_Cur->addr) &&
            (p_Cur->prev != NULL) )
      {
         p_Prev = p_Cur;
         p_Cur = p_Cur->prev;
      }

      if( (p_Cur->prev == NULL) &&
            (map.addr < p_Cur->addr) )
      {
         p_Cur->prev = p_Mess;
         p_Mess->next = p_Cur;
         p_Mess->prev = NULL;
         list->list_Map = p_Mess;
      }
      else
      {
         p_Mess->next = p_Prev;
         p_Mess->prev = p_Cur;
         p_Prev->prev = p_Mess;
         p_Cur->next = p_Mess;
      }
   }

   p_Mess->addr = map.addr;
   p_Mess->size = map.size;
   p_Mess->is_printed = Z_FALSE;
   strcpy(p_Mess->label, map.label);
   p_Mess->type = map.type;
   p_Mess->references.list_Ref = NULL;

   return(p_Mess);
}

/***********************************************************
 *  Function: LIST_GetMappingAtAddr     File: chainlist.c  *
 ***********************************************************
 * Usage:
 *
 * Gets the element of the list which is at address i_Addr.
 ***********************************************************/ 
Mapping *LIST_GetMappingAtAddr(LIST_Mapping *list, uint32 i_Addr)
{
   Mapping *p_CurMap;

   p_CurMap = list->list_Map;

   while(p_CurMap != NULL)
   {
      if(p_CurMap->addr == i_Addr)
         return(p_CurMap);
      p_CurMap = p_CurMap->next;
   }

   return(NULL);
}


/**************************************************************
 *                    Reference functions                     *
 **************************************************************/ 

/***********************************************************
 *  Function: LIST_FreeRef              File: chainlist.c  *
 ***********************************************************
 * Usage:
 *
 * Frees all elements of a list.
 ***********************************************************/ 
int32 LIST_FreeRef(LIST_Reference *list)
{
   Reference *p_Next,*p_Ref = list->list_Ref;

   if( p_Ref == NULL )
   {
      return(Z_OK);
   }

   while (p_Ref->next != NULL)
   {
      p_Next = p_Ref->next;
      free(p_Ref);
      p_Ref=p_Next;

   }

   free(p_Ref);
   list->list_Ref = NULL;

   return(Z_OK);
}

/***********************************************************
 *  Function: LIST_DeleteRef            File: chainlist.c  *
 ***********************************************************
 * Usage:
 *
 * Removes an element of the list.
 ***********************************************************/ 
int32 LIST_DeleteRef(LIST_Reference *list, int32 i_Index)
{
   int32 i=0;
   Reference *p_Ref, *p_Prev;

   if (i_Index == 0) 
   {
      p_Prev = NULL;
      p_Ref = list->list_Ref;
   }
   else 
   {
      p_Ref = list->list_Ref;
      while (i++ < i_Index-1)
      {
         p_Ref = p_Ref->next;
      }
      p_Prev = p_Ref;
      p_Ref = p_Ref->next;
   }

   if (p_Prev)
   {
      p_Prev->next = p_Ref->next;
   }
   else
   {
      list->list_Ref = p_Ref->next;
   }

   free(p_Ref);

   return(Z_OK);
}

/***********************************************************
 *  Function: LIST_AppendRef            File: chainlist.c  *
 ***********************************************************
 * Usage:
 *
 * Appends an element to the list.
 ***********************************************************/ 
int32 LIST_AppendRef(LIST_Reference *list, Reference ref)
{
   Reference *p_List,*p_Mess= list->list_Ref;

   if (list->list_Ref == NULL)
   {
      list->list_Ref = (Reference*) malloc(sizeof(Reference));
      list->list_Ref->addr = ref.addr;
      list->list_Ref->type = ref.type;
      list->list_Ref->next = NULL;
   }
   else
   {
      while ( p_Mess->next != NULL)
      {
         p_Mess = p_Mess->next;
      }

      p_List = (Reference*) malloc(sizeof(Reference));
      p_List->addr = ref.addr;
      p_List->type = ref.type;
      p_List->next = NULL;

      p_Mess->next = p_List;

   }

   return(Z_OK);
}

/**************************************************************
 *                    Reference functions                     *
 **************************************************************/ 

/***********************************************************
 *  Function: LIST_FreeCallees           File: chainlist.c  *
 ***********************************************************
 * Usage:
 *
 * Frees all elements of a list.
 ***********************************************************/ 
int32 LIST_FreeCallees(LIST_Callee *list)
{
   Callee *p_Next,*p_Ref;
   
   if(list == NULL)
      return(Z_OK);

   p_Ref = list->list_ref;
   if( p_Ref == NULL )
      return(Z_OK);

   while (p_Ref->next != NULL)
   {
      p_Next = p_Ref->next;
      free(p_Ref);
      p_Ref=p_Next;

   }

   free(p_Ref);
   list->list_ref = NULL;
   return(Z_OK);
}

/***********************************************************
 *  Function: LIST_AddCallee            File: chainlist.c  *
 ***********************************************************
 * Usage:
 *
 * Appends an element to the list.
 ***********************************************************/ 
int32 LIST_AddCallee(LIST_Callee *list, Mapping *mapping)
{
   Callee *p_List,*p_Mess;
   
   if(list == NULL)
      return(Z_OK);
   
   p_Mess = list->list_ref;

   if (list->list_ref == NULL)
   {
      list->list_ref = (Callee*) malloc(sizeof(Reference));
      list->list_ref->mapping = mapping;
      list->list_ref->next = NULL;
   }
   else
   {
      while ( p_Mess->next != NULL)
         p_Mess = p_Mess->next;

      p_List = (Callee*) malloc(sizeof(Reference));
      p_List->mapping = mapping;
      p_List->next = NULL;
      p_Mess->next = p_List;
   }

   return(Z_OK);
}

