/******************************************
 *      the Beos Reversing Engine         * 
 *              by Zadig                  *
 *                                        *
 * Custom symbols injection               *
 ******************************************/


/*these files must always be includes before all others*/
#include <types.h>
#include "globals.h"
#include "structs.h"
#include "reveng_errors.h"
/* end these files must always be includes before all others*/

#include <zem_errors.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <stdio.h>
#include <string.h>

#include "chainlist.h"


/***********************************************************
 *  Function: RE_GetCustomSymbols         File: custsym.c  *
 ***********************************************************
 * Usage:
 *
 * Reads the custom symbols files and adds them in the symbols
 * list.
 ***********************************************************/ 
int32 RE_GetCustomSymbols(Settings *projectSettings, LIST_Mapping *globalMapping)
{
   FILE *f_SymbolFile;
   LIST_Mapping *p_CurListMapping;
   uint8 sz_CurLine[STRING_MAX_SIZE], *psz_Message;
   int32 i_CurAddr,i;
   Mapping curMap, *p_TmpMap;
   uint32 i_CurMap;
   int32 i_ret;
   uint8 b_Found;

   DBG_PRINT("Entering RE_GetCustomSymbols...\n");    
   /* Initialisation */
   curMap.next = NULL;
   curMap.references.list_Ref = NULL;
   /* End Initialisation */

   f_SymbolFile = fopen(projectSettings->sym_FileName,"r"); 
   if(f_SymbolFile == NULL)
   {
      return(FILE_ERROR_OPEN);
   }

   for(i=0; i<2; i++)
   {
      /* goto begining of custom symbol table */
      do
      {
         i_ret = (uint32)fgets(sz_CurLine,STRING_MAX_SIZE,f_SymbolFile);
         if(i_ret == 0)
         {
            fclose(f_SymbolFile);
            return(Z_OK);
         }
      } while( (strncmp(sz_CurLine,CONF_SECTION_CUST_SYM, 
                  strlen(CONF_SECTION_CUST_SYM)) != 0) &&
            (strncmp(sz_CurLine,CONF_SECTION_CUST_DYN_SYM, 
                     strlen(CONF_SECTION_CUST_DYN_SYM)) != 0) );

      if(strncmp(sz_CurLine, CONF_SECTION_CUST_DYN_SYM,
                  strlen(CONF_SECTION_CUST_DYN_SYM)) == 0)
      {
         i_CurMap = MAP_DYNSYM;
         p_CurListMapping = &(globalMapping[MAP_DYNSYM]);
         i_ret = (uint32)fgets(sz_CurLine,STRING_MAX_SIZE,f_SymbolFile);
         if(i_ret == 0)
         {
            fclose(f_SymbolFile);
            return(Z_OK);
         }       
      }
      else
      {
         i_CurMap = MAP_SYMBOLS;
         p_CurListMapping = &(globalMapping[MAP_SYMBOLS]);
      }

      /* reads all symbols */
      i_ret = (uint32)fgets(sz_CurLine,STRING_MAX_SIZE,f_SymbolFile);
      while( (i_ret != 0) && (sz_CurLine[0] != '\n') )
      {
         psz_Message = sz_CurLine;

         if(*psz_Message == '#') /* comments */
         {
            i_ret = (uint32)fgets(sz_CurLine,STRING_MAX_SIZE,f_SymbolFile);
            continue;
         }
         
         /* addr of object */
         sscanf(sz_CurLine,"%X ",(int*)&(i_CurAddr));
         curMap.addr = (uint32) i_CurAddr;
         
         while( ((*psz_Message) != ' ') && ((*psz_Message) != '\n') )
            psz_Message++;
         if(*psz_Message == '\n')
         {
            i_ret = (uint32)fgets(sz_CurLine,STRING_MAX_SIZE,f_SymbolFile);
            continue;
         }
         while( ((*psz_Message) == ' ') && ((*psz_Message) != '\n') )
            psz_Message++;
         if(*psz_Message == '\n')
         {
            i_ret = (uint32)fgets(sz_CurLine,STRING_MAX_SIZE,f_SymbolFile);
            continue;
         }

         /* reads the type of the object */
         curMap.type = OBJ_TYPE_FUNCTION;
         if(psz_Message[0] == 'O')
            curMap.type = OBJ_TYPE_OBJECT;
         
         while( ((*psz_Message) != ' ') && ((*psz_Message) != '\n') )
            psz_Message++;
         if(*psz_Message == '\n')
         {
            i_ret = (uint32)fgets(sz_CurLine,STRING_MAX_SIZE,f_SymbolFile);
            continue;
         }
         while( ((*psz_Message) == ' ') && ((*psz_Message) != '\n') )
            psz_Message++;
         if(*psz_Message == '\n')
         {
            i_ret = (uint32)fgets(sz_CurLine,STRING_MAX_SIZE,f_SymbolFile);
            continue;
         }

         /* get symbol size */
         sscanf(psz_Message,"%X ",(int*)&curMap.size);

         while( ((*psz_Message) != ' ') && ((*psz_Message) != '\n') )
            psz_Message++;
         if(*psz_Message == '\n')
         {
            i_ret = (uint32)fgets(sz_CurLine,STRING_MAX_SIZE,f_SymbolFile);
            continue;
         }
         while( ((*psz_Message) == ' ') && ((*psz_Message) != '\n') )
            psz_Message++;
         if(*psz_Message == '\n')
         {
            i_ret = (uint32)fgets(sz_CurLine,STRING_MAX_SIZE,f_SymbolFile);
            continue;
         }

         /* label */
         strcpy(curMap.label,psz_Message);         
         curMap.label[strlen(curMap.label)-1] = '\0'; /* remove the '\n' */

         /* don't add it if an occurence already exists */
         p_TmpMap = p_CurListMapping->list_Map;
         b_Found = Z_FALSE;
         while( (p_TmpMap != NULL) &&
               (b_Found == Z_FALSE) )
         {
            if(p_TmpMap->addr == curMap.addr)
               b_Found = Z_TRUE;
            p_TmpMap = p_TmpMap->next;
         }

         if(b_Found == Z_FALSE)    
            LIST_AddMap(p_CurListMapping,curMap);

         i_ret = (uint32)fgets(sz_CurLine,STRING_MAX_SIZE,f_SymbolFile);
      } /* end while */
   } /* end for */

   fclose(f_SymbolFile);
   DBG_PRINT("End of RE_GetCustomSymbols\n");  
   return(Z_OK);
}

