/******************************************
 *      the Beos Reversing Engine         * 
 *              by Zadig                  *
 *                                        *
 * This file contains functions useful    *
 * to obtain "usable" files from objdump. *
 ******************************************/


/*these files must always be includes before all others*/
#include <types.h>
#include "globals.h"
#include "structs.h"
#include "reveng_errors.h"
/* end these files must always be includes before all others*/

#include <string.h>
#include "misc.h"
#include "graph.h"

/***********************************************************
 *  Function: RE_PrintMappingCallee         File: graph.c  *
 ***********************************************************
 * Usage:
 *
 * Print the mapping label and all its callees.
 ***********************************************************/ 
int32 RE_PrintMappingCallee(FILE* f_OutputFile, uint32 i_Level, 
                            Mapping *p_Mapping)
{
   if(i_Level < 5)
   {
      if(p_Mapping->is_printed == Z_TRUE) /* recursive call */
      {
         fprintf(f_OutputFile, "%s+recursive call to %s\n", 
               RE_GetLevelStr(i_Level), p_Mapping->label);
         return(Z_OK);
      }
      else
      {
         fprintf(f_OutputFile, "%s+%s\n", RE_GetLevelStr(i_Level), 
               p_Mapping->label);
      }
      p_Mapping->is_printed = Z_TRUE;
      if(p_Mapping->callees.list_ref != NULL)
         RE_PrintGraphEntry(f_OutputFile, i_Level+1, p_Mapping->callees.list_ref);
      p_Mapping->is_printed = Z_FALSE;
   }
   return(Z_OK);
}

/***********************************************************
 *  Function: RE_PrintfGraphEntry           File: graph.c  *
 ***********************************************************
 * Usage:
 *
 * Gets alls references of objects that are in the mapping.
 ***********************************************************/ 
int32 RE_PrintGraphEntry(FILE *f_OutputFile, uint32 i_Level, Callee *p_Entry)
{
   while(p_Entry != NULL)
   {
      RE_PrintMappingCallee(f_OutputFile, i_Level, p_Entry->mapping);
      p_Entry = p_Entry->next;
   }
   return(Z_OK);
}

/***********************************************************
 *  Function: RE_PrintfGraph                File: graph.c  *
 ***********************************************************
 * Usage:
 *
 * Gets alls references of objects that are in the mapping.
 ***********************************************************/ 
int32 RE_PrintGraph(FILE *f_OutputFile, LIST_Mapping *projectMapping)
{
   Mapping *p_CurFunction;
   uint8 *psz_FunctionName;

   DBG_PRINT("Entering RE_PrintGraph...\n");     
   
   if(projectMapping == NULL)
      return(RE_ERROR_PARAMETER);
   
   p_CurFunction = projectMapping->list_Map;
   while(p_CurFunction != NULL)
   {
      /* don't process templates */
      psz_FunctionName = strstr(p_CurFunction->label,"<"); 
      
      if( (psz_FunctionName == NULL) &&      /* templates contain '<' and '>' */
          ( (p_CurFunction->type == OBJ_TYPE_CLASS) ||
            (p_CurFunction->type == OBJ_TYPE_FUNCTION) ) &&
          (strstr(p_CurFunction->label, RTTI_SYMBOL) == NULL) 
        )
      {
         if(p_CurFunction->callees.list_ref != NULL)
         {
            RE_PrintMappingCallee(f_OutputFile, 0, p_CurFunction);
            fprintf(f_OutputFile, "\n");
         }
      }
      p_CurFunction = p_CurFunction->next;
   }

   DBG_PRINT("End of RE_PrintGraph\n");
   return(Z_OK);
}

