/**********************************************
 *          the Beos Reversing Engine         * 
 *                  by Zadig                  *
 *                                            *
 * This file contains all init functions.     *
 **********************************************/

/*these files must always be includes before all others*/
#include <types.h>
#include "globals.h"
#include "structs.h"
#include "reveng_errors.h"
/* end these files must always be includes before all others*/

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "chainlist.h"
#include "misc.h"
#include "init.h"

/***********************************************************
 *  Function: RE_ReadCmdLine                 File: init.c  *
 ***********************************************************
 * Usage:
 * Reads the command line.
 *
 *  Reads the command line parameters passed to RevEng.
 *
 * NOTE: To debug with bdb, define RE_BDB_ is necessary.
 *       It replaces the command line with other values
 *       (see in code).
 ***********************************************************/ 
int32 RE_ReadCmdLine(int16 argc, int8 *argv[], Settings *projectSettings)
{
   uint16 i_CurArg;

   projectSettings->splitOutput  = Z_FALSE;
   projectSettings->printPlt     = Z_FALSE; 
   projectSettings->printDsym    = Z_FALSE;
   projectSettings->ctags        = Z_FALSE;
   projectSettings->offsetSort   = 0;
   projectSettings->printGraph   = Z_FALSE;
   projectSettings->symInject    = Z_FALSE;
   strcpy(projectSettings->blank_line,"\n");    /* default is 1 blank line */
   projectSettings->sym_FileName[0] = '\0';

   if(argc <= 1)
      return(RE_ERROR_PARAMETER);

   i_CurArg = 1;
   while( (i_CurArg < argc) && (argv[i_CurArg][0] == '-') )
   {
      if(strcmp((char*)(argv[i_CurArg]),"-split") == 0)
         projectSettings->splitOutput = Z_TRUE;
      else if(strcmp((char*)(argv[i_CurArg]),"-plt") == 0)
         projectSettings->printPlt = Z_TRUE;
      else if(strcmp((char*)(argv[i_CurArg]),"-ctags") == 0)
         projectSettings->ctags = Z_TRUE;
      else if(strcmp((char*)(argv[i_CurArg]),"-nobl") == 0)
         projectSettings->blank_line[0] = '\0';
      else if(strcmp((char*)(argv[i_CurArg]),"-dsym") == 0)
         projectSettings->printDsym = Z_TRUE;
      else if(strcmp((char*)(argv[i_CurArg]),"-nodupsym") == 0)
         projectSettings->noDupSym = Z_TRUE;
      else if(strcmp((char*)(argv[i_CurArg]),"-graph") == 0)
         projectSettings->printGraph = Z_TRUE;
      else if(strstr((char*)(argv[i_CurArg]),"-osort=") != NULL)
      {
         char *psz_CurType = (char*)argv[i_CurArg];

         psz_CurType += 7;
         while(*psz_CurType != '\0')
         {
            if(*psz_CurType == 'd')
            {
               psz_CurType++;
               if(*psz_CurType == 'f')
               {
                  projectSettings->offsetSort |= OSORT_DFUNCTIONS;
                  psz_CurType++;
               }
               /* offset class sorting disabled for the moment */
               /*
               else if(*psz_CurType == 'c')
               {
                  projectSettings->offsetSort |= OSORT_DCLASSES;
                  psz_CurType++;
               }
               */
            }
            else if(*psz_CurType == 'f')
            {
               projectSettings->offsetSort |= OSORT_FUNCTIONS;
               psz_CurType++;
            }
            /* offset class sorting disabled for the moment */
            /*
            else if(*psz_CurType == 'c')
            {
               projectSettings->offsetSort |= OSORT_CLASSES;
               psz_CurType++;
            }
            */
            else if(*psz_CurType == 'o')
            {
               projectSettings->offsetSort |= OSORT_OBJECTS;
               psz_CurType++;
            }
            else if(*psz_CurType == 's')
            {
               projectSettings->offsetSort |= OSORT_STRINGS;
               psz_CurType++;
            }
         }
      }
      else if(strcmp((char*)(argv[i_CurArg]),"-symtab") == 0)
      {
         i_CurArg++;
         strcpy(projectSettings->sym_FileName, (char*)(argv[i_CurArg]));
         projectSettings->symInject = Z_TRUE;
      }
      else
         return(RE_ERROR_PARAMETER);
      i_CurArg++;
   }

   /* The 2 last params must be the target's name and
      the destination's file name. */
   if( (argc - i_CurArg) != 2)
      return(RE_ERROR_PARAMETER);

   strcpy((char*)(projectSettings->exe_FileName), (char*)(argv[i_CurArg]));
   strcpy((char*)(projectSettings->objdump_FileName), (char*)(argv[i_CurArg]));
   strcat((char*)(projectSettings->objdump_FileName),".objdmp");

   i_CurArg++;
   strcpy((char*)(projectSettings->dest_FileName), (char*)(argv[i_CurArg]));  
   strcpy((char*)(projectSettings->asm_SrcFileName), (char*)(argv[i_CurArg]));
   strcat((char*)(projectSettings->asm_SrcFileName),".asmsrc");
   
   /*debugging infromations*/  
   DBG_PRINT("Call to RE_ReadCmdLine:\n");
   DBG_PRINT("exe_FileName == ");
   DBG_PRINT((char*)(projectSettings->exe_FileName));
   DBG_PRINT("\n");

   DBG_PRINT("dest_FileName == ");
   DBG_PRINT((char*)(projectSettings->dest_FileName));
   DBG_PRINT("\n");

   DBG_PRINT("objdump_FileName == ");
   DBG_PRINT((char*)(projectSettings->objdump_FileName));
   DBG_PRINT("\n");

   DBG_PRINT("asm_srcFileName == ");
   DBG_PRINT((char*)(projectSettings->asm_SrcFileName));
   DBG_PRINT("\n");
   DBG_PRINT("\n");    
   return(Z_OK);
}

/***********************************************************
 *  Function: RE_ReadConfigFile              File: init.c  *
 ***********************************************************
 * Usage:
 *
 *  Reads the configuration file and fills some parts of
 * the globalMapping variable.
 *  - Reads messages codes.
 *  - This is where params size will be read when
 *    implemented.
 ***********************************************************/ 
int32 RE_ReadConfigFile(LIST_Mapping *globalMapping)
{
   uint8    sz_ConfigFile[STRING_MAX_SIZE];
   int32 i_ret;
   FILE *f_ConfigFile;

   sprintf(sz_ConfigFile, "%s%s", getenv("HOME"), CONFIG_FILE);   
   f_ConfigFile = fopen(sz_ConfigFile, "r"); 
   if(f_ConfigFile == NULL)
      return(FILE_ERROR_OPEN);

   i_ret = RE_ReadMessages(f_ConfigFile, &(globalMapping[MAP_MESSAGES]) );
   if(i_ret != Z_OK)
   {
      fclose(f_ConfigFile);
      return(i_ret);
   }

   i_ret = RE_ReadParams(f_ConfigFile, &(globalMapping[MAP_PARAMS]) );
   if(i_ret != Z_OK)
   {
      fclose(f_ConfigFile);
      return(i_ret);
   }

   fclose(f_ConfigFile);
   return(Z_OK);
}

/***********************************************************
 *  Function: RE_ReadMessages                File: init.c  *
 ***********************************************************
 * Usage:
 *
 * Reads messages codes in the configuration file.
 ***********************************************************/ 
int32 RE_ReadMessages(FILE* f_ConfigFile, LIST_Mapping *messagesList)
{
   int32 i_ret;
   Mapping curMap;
   uint8 sz_CurLine[STRING_MAX_SIZE];  
   uint8 *psz_Message;
   uint16 i_Index;

   fseek(f_ConfigFile, 0, SEEK_SET);

   /* goto messages section */  
   i_ret = (uint32)fgets((char*)(sz_CurLine), STRING_MAX_SIZE, f_ConfigFile);
   while( (i_ret != 0) &&
         (strncmp((char*)(sz_CurLine),"[Messages]",10) != 0) )
   {
      i_ret = (uint32)fgets((char*)(sz_CurLine), STRING_MAX_SIZE, f_ConfigFile);
   }

   if(i_ret == 0)
      return(Z_OK);

   /* read messages */
   i_ret = (uint32)fgets((char*)(sz_CurLine),STRING_MAX_SIZE,f_ConfigFile);
   while( (i_ret != 0) &&
         (sz_CurLine[0] != '\n') )
   {
      /* read message label */
      i_Index = 0;
      psz_Message = sz_CurLine;
      while((*psz_Message) != '=')
      {
         curMap.label[i_Index] = *psz_Message;
         psz_Message++;
         i_Index++;
      }
      curMap.label[i_Index] = '\0';
      psz_Message += 2;

      /* then read message value */        
      while(*psz_Message != '\'')
      {
         curMap.addr <<= 8;
         curMap.addr |= (*psz_Message);
         psz_Message++;
      }

      curMap.type = OBJ_TYPE_MESSAGE;
      LIST_AddMap(messagesList,curMap);

      i_ret = (uint32)fgets((char*)(sz_CurLine),STRING_MAX_SIZE,f_ConfigFile);
   }

   return(Z_OK);
}

/***********************************************************
 *  Function: RE_ReadParams                  File: init.c  *
 ***********************************************************
 * Usage:
 *
 * Reads parameters size in the configuration file.
 ***********************************************************/ 
int32 RE_ReadParams(FILE* f_ConfigFile, LIST_Mapping *ParamsList)
{
   int32 i_ret;
   Mapping curMap;
   uint8 sz_CurLine[STRING_MAX_SIZE];  
   uint8 *psz_Param;
   uint16 i_Index;

   fseek(f_ConfigFile, 0, SEEK_SET);

   /* goto params section */  
   i_ret = (uint32)fgets((char*)(sz_CurLine), STRING_MAX_SIZE, f_ConfigFile);
   while( (i_ret != 0) &&
         (strncmp((char*)(sz_CurLine),"[Parameters]",8) != 0) )
   {
      i_ret = (uint32)fgets((char*)(sz_CurLine), STRING_MAX_SIZE, f_ConfigFile);
   }

   if(i_ret == 0)
      return(Z_OK);

   /* reads params size */
   i_ret = (uint32)fgets((char*)(sz_CurLine),STRING_MAX_SIZE,f_ConfigFile);
   while( (i_ret != 0) && (sz_CurLine[0] != '\n') )
   {
      /* read parameter label */
      i_Index = 0;
      psz_Param = sz_CurLine;
      while((*psz_Param) != '=')
      {
         curMap.label[i_Index] = *psz_Param;
         psz_Param++;
         i_Index++;
      }
      curMap.label[i_Index] = '\0';
      psz_Param++;

      /* then read parameter size */        
      sscanf((char*)(psz_Param), "%u", (unsigned int*)(&curMap.addr) );    
      curMap.type = OBJ_TYPE_PARAM;
      LIST_AddMap(ParamsList,curMap);

      i_ret = (uint32)fgets((char*)(sz_CurLine),STRING_MAX_SIZE,f_ConfigFile);
   }

   return(Z_OK);
}

