/****************************************
 *      the Beos Reversing Engine       * 
 *              by Zadig                *
 *                                      *
 * This file is the code enrty point    *
 * of this soft.                        *
 ****************************************/

/*these files must always be includes before all others*/
#include <types.h>
#include "globals.h"
#include "structs.h"
#include "reveng_errors.h"
/* end these files must always be includes before all others*/

#include <zem_errors.h>
#include <unistd.h>
#include <stdio.h>
#include <time.h>

#include "init.h"
#include "misc.h"
#include "reveng.h"
#include "objdump.h"
#include "asm.h"
#include "output.h"
#include "chainlist.h"

/***********************************************************
 *  Function: main                           File: main.c  *
 ***********************************************************
 * Usage:
 *
 *  This is the entry point of the program. It is the
 * place where are defined all project variables.
 ***********************************************************/ 
int main(int argc, char *argv[])
{
   time_t   startTime, endTime;
   Settings projectSettings;
   LIST_Mapping globalMapping[MAP_MAX];
   PLT_Jump projectPLT[MAX_PLT_ENTRIES];
   int32 i_ret;
   uint32 i;
   uint8  sz_Header[255];

   DBG_START;
   DBG_PRINT("BeOS Reveng\n");
   DBG_PRINT("Starting Session.\n\n");

   /* Initialisation */
   for(i=0; i< MAP_MAX; i++)
   {
      globalMapping[i].list_Map = NULL;
      globalMapping[i].tail = NULL;
      globalMapping[i].alpha = NULL;
   }
   for(i=0; i< MAX_PLT_ENTRIES; i++)
   {
      projectPLT[i].addr = 0;
   }
   /* End Initialisation */  
   RE_GetHeader(sz_Header);
   printf("%s", sz_Header);

   i_ret = RE_ReadCmdLine((int16)argc, (int8**)argv, &projectSettings);
   if(i_ret == RE_ERROR_PARAMETER)
   {
      RE_PrintHelp();
      return (i_ret);
   }

   i_ret = RE_ReadConfigFile((LIST_Mapping*)globalMapping);
   if(i_ret != Z_OK)
   {
      if(i_ret == FILE_ERROR_OPEN)
         printf("Warning: Config file not found.\n");
      else
      {
         printf("Error while reading config file!\n");
         return(i_ret);
      }
   }
   /* removes tmp files if they exists */
   unlink(projectSettings.objdump_FileName);
   unlink(projectSettings.asm_SrcFileName);  

   startTime = time(NULL);
   /* disassembles the file */
   printf("Disassembling file...\n");
   i_ret = RE_DisassembleFile(&projectSettings);
   if(i_ret != Z_OK)
   {
      printf("Error while disassembling file!\n");     
      return(i_ret);
   }

   /* reads its sections properties*/
   i_ret = RE_GetObjdumpSections(&projectSettings, projectSettings.sections);
   if(i_ret != Z_OK)
   {
      printf("Error while reading sections!\n");     
      return(i_ret);    
   }

   if(projectSettings.symInject == Z_TRUE)
   {
      printf("Reading custom symbol tables...\n");
      i_ret = RE_GetCustomSymbols(&projectSettings, 
                                 (LIST_Mapping*)globalMapping);
      if(i_ret != Z_OK)
      {
         printf("Error while reading custom symbol tables!\n");     
         return(i_ret);
      }
   }

   /* retrieves infos from disassembly */
   printf("Building mapping of file...\n");
   i_ret = RE_GetSymbolsMapping(&projectSettings, (LIST_Mapping*)globalMapping);
   if(i_ret != Z_OK)
   {
      printf("Error while reading mapping in Objdump file!\n");     
      return(i_ret);
   }

   /* creates a clean asm file */
   printf("Creating clean disassembly file...\n");  
   i_ret = ASM_CreateAsmFile(&projectSettings);
   if(i_ret != Z_OK)
   {
      printf("Error while creating asm source file!\n");     
      return(i_ret);    
   }

   /* get dynamic symbols real call addresses */
   printf("Getting symbols references...\n");
   i_ret = RE_GetPLTJumps(&projectSettings, projectPLT);
   if(i_ret == Z_OK)
   {
      i_ret = RE_GetDynSymsRealAddr((LIST_Mapping*)globalMapping, projectPLT);
      if(i_ret != Z_OK)
      {
         printf("Error while reading real address of dynamic symbols!\n");     
         return(i_ret);    
      }
   }
   else
   {
      printf("Error while reading the plt section!\n"
             "Continuing without dynamic symbols references\n");
   }

   /* looks for all references */
   i_ret = RE_GetMappingReferences(&projectSettings, (LIST_Mapping*)globalMapping);
   if(i_ret != Z_OK)
   {
      printf("Error while getting references of functions/objects!\n");     
      return(i_ret);    
   }

   unlink(projectSettings.objdump_FileName);
   printf("Creating output file...\n");
   i_ret = RE_CreateOutputFile(&projectSettings, (LIST_Mapping*)globalMapping);
   if(i_ret != Z_OK)
   {
      printf("Error while creating output file!\n");     
      return(i_ret);    
   }

   unlink(projectSettings.asm_SrcFileName);
   for(i=0; i< MAP_MAX; i++)
   {
      LIST_FreeMap(&(globalMapping[i]));
   }

   endTime = time(NULL); 
   printf("Disassembly completed in %dmins %dsecs.\n", (int) ((endTime - startTime) / 60),
         (int) ((endTime - startTime) % 59)); 

   DBG_PRINT("End of session.\n");
   DBG_END;
   return Z_OK;
}
