/******************************************
 *      the Beos Reversing Engine         * 
 *              by Zadig                  *
 *                                        *
 * This file contains functions useful    *
 * to obtain "usable" files from objdump. *
 ******************************************/


/*these files must always be includes before all others*/
#include <types.h>
#include "globals.h"
#include "structs.h"
#include "reveng_errors.h"
/* end these files must always be includes before all others*/

#include <zem_errors.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <sys/wait.h>   /* for solaris */

#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "objdump.h"
#include "chainlist.h"

/***********************************************************
 *  Function: RE_DisassembleFile          File: objdump.c  *
 ***********************************************************
 * Usage:
 *
 * Disassemble a file by using objdump.
 ***********************************************************/ 
int32 RE_DisassembleFile(Settings *fileSettings)
{
   int pfils;
   int status;
   int fd, fd_old;
   uint8 sz_FileName[STRING_MAX_SIZE];
   uint8 sz_CmdLine[STRING_MAX_SIZE];
   int16 w;

   strcpy(sz_FileName,fileSettings->objdump_FileName);
   sprintf(sz_CmdLine," -d --demangle -T -x -R %s",fileSettings->exe_FileName);

   DBG_PRINT("\nRE_DisassembleFile: ");
   DBG_PRINT(sz_CmdLine);
   DBG_PRINT("\n");  

   /* redirects stdout to filesettigs->objdump_FileName */
   fd = open(fileSettings->objdump_FileName, O_RDWR | O_CREAT);    
   fd_old = dup(1);
   dup2(fd,1);

   /* launches objdump */
   if((pfils = fork()) < 0)
   {
      return(RE_ERROR_FORK);
   }
   else if(pfils == 0)
   {
      execlp("objdump","objdump","-d",
            "--demangle","-T","-x", "-R",fileSettings->exe_FileName,NULL);
      exit(-1);
   }
   while((w = wait(&status)) != pfils && w != -1);

   /* restores stdout */
   dup2(fd_old, 1);
   close(fd_old);
   close(fd);

   /* ensure that we can read it later */
   chmod(fileSettings->objdump_FileName, S_IRWXU);

   return(Z_OK);
}

/***********************************************************
 *  Function: RE_GetObjdumpSections       File: objdump.c  *
 ***********************************************************
 * Usage:
 *
 * Retrieves the sections addresses and sizes.
 ***********************************************************/ 
int32 RE_GetObjdumpSections(Settings *projectSettings, Section *projectSections)
{
   FILE *f_DisasmFile;
   uint8 sz_CurLine[STRING_MAX_SIZE], *psz_Section;  
   uint8 sz_CurSection[STRING_MAX_SIZE];
   int32 i_CurSectionNum,i_CurAddr,i_CurSize;
   int32 i_ret,i;

   DBG_PRINT("Entering RE_GetObjdumpSections...\n");  
   /* Initialisation */
   i_CurSectionNum = 0;

   for(i=0; i< MAX_SECTIONS; i++)
   {
      projectSections[i].name[0] = '\0';
   }
   /* End Initialisation */

   f_DisasmFile = fopen(projectSettings->objdump_FileName,"r"); 
   if(f_DisasmFile == NULL)
   {
      return(FILE_ERROR_OPEN);
   }

   /* goto begining of sections definition */
   do
   {
      i_ret = (uint32)fgets(sz_CurLine,STRING_MAX_SIZE,f_DisasmFile);
      if(i_ret == 0)
      {
         fclose(f_DisasmFile);
         return(FILE_ERROR_READ);
      }
   } while(strncmp(sz_CurLine,"Sections:",9) != 0) ;

   fgets(sz_CurLine,STRING_MAX_SIZE,f_DisasmFile);
   fgets(sz_CurLine,STRING_MAX_SIZE,f_DisasmFile);

   /* reads all sections */   
   while( (sz_CurLine[0] == ' ') && (i_CurSectionNum < MAX_SECTIONS) )
   {

      /* reads section name */
      i = 0;
      psz_Section = sz_CurLine + 4;
      while(*psz_Section != ' ')
      {
         sz_CurSection[i] = *psz_Section;
         psz_Section++;
         i++;
      }
      sz_CurSection[i] = '\0';

      /* reads section size and addr*/
      while(*psz_Section == ' ')
      {
         psz_Section++;
      }
      sscanf(psz_Section, "%X %X",(unsigned int*) &i_CurSize, (unsigned int*)&i_CurAddr);     

      strcpy(projectSections[i_CurSectionNum].name, sz_CurSection);
      projectSections[i_CurSectionNum].addr = i_CurAddr;
      projectSections[i_CurSectionNum].size = i_CurSize;     

      /* goto next section */
      fgets(sz_CurLine,STRING_MAX_SIZE,f_DisasmFile);
      i_ret = (uint32) fgets(sz_CurLine,STRING_MAX_SIZE,f_DisasmFile);   
      if(i_ret == 0)
      {
         fclose(f_DisasmFile);
         return(FILE_ERROR_READ);
      }
      i_CurSectionNum++;

   }
   if(i_CurSectionNum == MAX_SECTIONS)
   {
      DBG_PRINT("Warning: too much sections in file\n");  
   }


   fclose(f_DisasmFile);
   DBG_PRINT("End of RE_GetObjdumpSections\n");  
   return(Z_OK);
}

/***********************************************************
 *  Function: RE_GetSymbolsMapping        File: objdump.c  *
 ***********************************************************
 * Usage:
 *
 * Builds the mapping of the functions/strings/objects of
 * an executable by using the objdump generated file.
 ***********************************************************/ 
int32 RE_GetSymbolsMapping(Settings *projectSettings, LIST_Mapping *globalMapping)
{
   FILE *f_DisasmFile, *f_ExeFile;
   LIST_Mapping *p_CurListMapping;
   uint8 sz_CurLine[STRING_MAX_SIZE], *psz_Label;  
   int32 i_CurAddr,i;
   Mapping curMap, *p_TmpMap;
   uint32 i_CurMap;
   int32 i_ret;
   uint8 b_Found;

   DBG_PRINT("Entering RE_GetObjdumpMapping...\n");    
   /* Initialisation */
   curMap.next = NULL;
   curMap.references.list_Ref = NULL;
   /* End Initialisation */

   f_DisasmFile = fopen(projectSettings->objdump_FileName,"r"); 
   if(f_DisasmFile == NULL)
   {
      return(FILE_ERROR_OPEN);
   }
   
   f_ExeFile =  fopen((char*)(projectSettings->exe_FileName),"r");  
   if(f_ExeFile == NULL)
   {
      fclose(f_DisasmFile);
      return(FILE_ERROR_OPEN);
   }

   for(i=0; i<3; i++)
   {
      /* goto begining of symbol table */
      do
      {
         i_ret = (uint32)fgets(sz_CurLine,STRING_MAX_SIZE,f_DisasmFile);
         if(i_ret == 0)
         {
            fclose(f_ExeFile);
            fclose(f_DisasmFile);
            return(Z_OK);
         }
      } while( (strncmp(sz_CurLine,"DYNAMIC SYMBOL TABLE:",13) != 0) &&
            (strncmp(sz_CurLine,"SYMBOL TABLE:",13) != 0) &&
            (strncmp(sz_CurLine,"DYNAMIC RELOCATION RECORDS",13) != 0) );

      if(strncmp(sz_CurLine,"DYNAMIC RELOCATION RECORDS",13) == 0)
      {
         i_CurMap = MAP_DYNSYM;
         p_CurListMapping = &(globalMapping[MAP_DYNSYM]);
         i_ret = (uint32)fgets(sz_CurLine,STRING_MAX_SIZE,f_DisasmFile);
         if(i_ret == 0)
         {
            fclose(f_ExeFile);
            fclose(f_DisasmFile);
            return(FILE_ERROR_READ);
         }       
      }
      else
      {
         i_CurMap = MAP_SYMBOLS;
         p_CurListMapping = &(globalMapping[MAP_SYMBOLS]);
      }

      /* reads all symbols */
      i_ret = (uint32)fgets(sz_CurLine,STRING_MAX_SIZE,f_DisasmFile);
      if(i_ret == 0)
      {
         fclose(f_ExeFile);
         fclose(f_DisasmFile);
         return(FILE_ERROR_READ);
      }       

      while( (sz_CurLine[0] != '\n') && (sz_CurLine[0] != '\0') )
      {
         if((strncmp(sz_CurLine,"00000000",8) != 0) &&
               (sz_CurLine[15] != ' ') )    
         {
            if( (i_CurMap == MAP_DYNSYM) && 
                  (strncmp(sz_CurLine+9,"R_386_JUMP_SLOT",15) != 0) )
            {
               i_ret = (uint32)fgets(sz_CurLine,STRING_MAX_SIZE,f_DisasmFile);
               if(i_ret == 0)
               {
                  fclose(f_ExeFile);
                  fclose(f_DisasmFile);
                  return(FILE_ERROR_READ);
               }       
               continue;
            }        

            /* reads the type of the object */
            curMap.type = OBJ_TYPE_FUNCTION;
            if(sz_CurLine[15] == 'O')
               curMap.type = OBJ_TYPE_OBJECT;

            /* addr of object */
            sscanf(sz_CurLine,"%X ",(int*)&(i_CurAddr));
            curMap.addr = (uint32) i_CurAddr;

            /* label of object */
            if(i_CurMap == MAP_DYNSYM)
            {
               psz_Label = &(sz_CurLine[27]);
            }
            else
            {
               psz_Label = &(sz_CurLine[17]);
               while(*psz_Label != '\t')  
                  psz_Label++;
               psz_Label++;
               /* get symbol size */
               sscanf(psz_Label,"%X ",(int*)&curMap.size);
               psz_Label += 9;
               while(*psz_Label == ' ')  /* strip symbol name */
                  psz_Label++;
            }

            if(strlen(psz_Label) > 2)
            {
               strcpy(curMap.label,psz_Label);
               /* remove the '\n' */
               curMap.label[strlen(psz_Label)-1] = '\0';
              
               /* don't add it if an occurence already exists */
               p_TmpMap = p_CurListMapping->list_Map;
               b_Found = Z_FALSE;
               while( (p_TmpMap != NULL) &&
                     (b_Found == Z_FALSE) )
               {
                  if(p_TmpMap->addr == curMap.addr)
                  {
                     if(projectSettings->noDupSym == Z_TRUE)
                     {
                        b_Found = Z_TRUE;
                     }
                     else if(strcmp(p_TmpMap->label,curMap.label) == 0)
                     {
                        b_Found = Z_TRUE;
                     }
                  }
                  p_TmpMap = p_TmpMap->next;
               }

               if(b_Found == Z_FALSE)    
               {
                  LIST_AddMap(p_CurListMapping,curMap);
               }
            } /* end if(strlen(psz_label) > 1) */

         }

         i_ret = (uint32)fgets(sz_CurLine,STRING_MAX_SIZE,f_DisasmFile);
         if(i_ret == 0)
         {
            fclose(f_ExeFile);
            fclose(f_DisasmFile);
            return(FILE_ERROR_READ);
         }       
      } /* end while */
   } /* end for */

   fclose(f_ExeFile);
   fclose(f_DisasmFile);
   DBG_PRINT("End of RE_GetObjdumpMapping\n");  
   return(Z_OK);
}

/***********************************************************
 *  Function: RE_DecomposeInstruction      File: reveng.c  *
 ***********************************************************
 * Usage:
 *
 * Decompose an objdump code line to its componnents.
 ***********************************************************/ 
int32 RE_DecomposeInstruction(uint8 *sz_AllLine, uint8 *sz_Addr, 
      uint8 *sz_HexCode, uint8 *sz_Instruction)
{
   uint8 *psz_CurElement;
   uint16 i_Index;

   /* gets addr of instruction */
   psz_CurElement = sz_AllLine;
   i_Index = 0;
   while(*psz_CurElement == ' ')
   {
      psz_CurElement++;        
      sz_Addr[i_Index] = '0';
      i_Index++;
   }
   while(*psz_CurElement  != ':')
   {
      sz_Addr[i_Index] = *psz_CurElement;
      i_Index++;
      psz_CurElement++;
   }
   sz_Addr[i_Index] = '\0';

   /* gets hex code of instruction */
   psz_CurElement +=2;
   i_Index = 0;
   while( (*psz_CurElement != '\n') &&
         (*psz_CurElement != '\0') &&
         (!( (*psz_CurElement == ' ') && 
             ((*(psz_CurElement+1) == ' ') || (*(psz_CurElement+1) == '\t') )
           ) )  )
   {
      if(*psz_CurElement != ' ')
      {
         sz_HexCode[i_Index] = *psz_CurElement;
         i_Index++;
      }
      psz_CurElement++;
   }
   sz_HexCode[i_Index] = '\0';

   /* if end of line, then return */
   if( (*psz_CurElement == '\n') ||
         (*psz_CurElement == '\0') )
   {
      sz_Instruction[0] = '\0';
      return(Z_OK);
   }

   /* gets label of instruction */
   i_Index = 0;
   psz_CurElement += 1;
   while(*psz_CurElement  ==' ')
   {
      psz_CurElement++;
   }
   psz_CurElement++;
   while( (*psz_CurElement != '<') && 
         (*psz_CurElement != '\n') &&
         (*psz_CurElement != '\0') )
   {
      sz_Instruction[i_Index] = *psz_CurElement;
      psz_CurElement++;
      i_Index++;
   }
   sz_Instruction[i_Index] = '\0';

   return(Z_OK);
}
