/**********************************************
 *          the Beos Reversing Engine         * 
 *                  by Zadig                  *
 *                                            *
 * This file contains specific disassembling  *
 * functions.                                 *
 **********************************************/

/*these files must always be includes before all others*/
#include <types.h>
#include "globals.h"
#include "structs.h"
#include "reveng_errors.h"
/* end these files must always be includes before all others*/

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "reveng.h"
#include "asm.h"
#include "chainlist.h"
#include "misc.h"


/***********************************************************
 *  Function: RE_GetPLTJumps               File: reveng.c  *
 ***********************************************************
 * Usage:
 *
 * Analyses the PLT section and build a map of it.
 ***********************************************************/ 
int32 RE_GetPLTJumps(Settings *projectSettings, PLT_Jump *projectPLT)
{
   FILE *f_AsmFile;  
   uint32 i_CurAddr, i_PLTIndex, i_CurPLT, i_CurValue;
   uint32 i_BaseAddr;
   int32 i_ret;
   uint32 i;
   uint8 sz_CurLine[STRING_MAX_SIZE];
   uint8 *psz_Inst;

   DBG_PRINT("Entering RE_GetPLTJumps...\n");    
   /* initialisation*/
   i_CurPLT = 0;

   /* clears PLT Jumps*/
   for(i=0; i < MAX_PLT_ENTRIES; i++)
   {
      projectPLT[i].addr = 0;
   }

   /* looks for .got section starting addr */
   i_BaseAddr = MAX_SECTIONS;
   for(i=0; i<MAX_SECTIONS; i++)
   {
      if(strcmp((char*)(projectSettings->sections[i].name),".got") == 0)
      {
         i_BaseAddr = projectSettings->sections[i].addr;
         break;
      }
   }
   if(i_BaseAddr == MAX_SECTIONS)
      return(Z_OK);

   /* end initialisation */

   /* searches for plt section */
   i_PLTIndex = MAX_SECTIONS;
   for(i=0; i< MAX_SECTIONS; i++)
   {
      if( strcmp((char*)(projectSettings->sections[i].name),".plt") == 0)
      {
         i_PLTIndex = i;
         break;
      }
   }

   if(i_PLTIndex == MAX_PLT_ENTRIES)
   {
      return(RE_ERROR_SECTION_NOT_FOUND);
   }

   f_AsmFile =  fopen((char*)(projectSettings->asm_SrcFileName),"r"); 
   if(f_AsmFile == NULL)
   {
      return(FILE_ERROR_OPEN);
   }

   /* goto begining of section */
   do
   {
      i_ret = (uint32)fgets((char*)(sz_CurLine),STRING_MAX_SIZE,f_AsmFile);
      if(i_ret == 0)
      {
         fclose(f_AsmFile);
         return(FILE_ERROR_READ);
      }
      sscanf((char*)(sz_CurLine),"%X",(unsigned int*)&i_CurAddr);
   } while(i_CurAddr != projectSettings->sections[i_PLTIndex].addr);

   /* gets all jumps in section (relative to ebx)*/
   do
   {
      /* goto instruction label */
      psz_Inst = sz_CurLine;
      for(i=0; i<2; i++)
      {
         while(*psz_Inst != ' ')
         {
            psz_Inst++;
         }
         while(*psz_Inst == ' ')
         {
            psz_Inst++;
         }
      }

      if(strncmp((char*)psz_Inst,"jmp    *0x",10) == 0)
      {
         sscanf((char*)(psz_Inst+10),"%X",(unsigned int*)&i_CurValue);
         projectPLT[i_CurPLT].addr = i_CurAddr;
         if(strstr(psz_Inst, "(%ebx)") != NULL) /* ebx based jump */
            projectPLT[i_CurPLT].value = i_CurValue + i_BaseAddr;
         else  /* absolute jump */
            projectPLT[i_CurPLT].value = i_CurValue;
         i_CurPLT++;
      }

      /* reads next line */
      i_ret = (uint32)fgets((char*)sz_CurLine,STRING_MAX_SIZE,f_AsmFile);
      if(i_ret == 0)
      {
         fclose(f_AsmFile);
         return(FILE_ERROR_READ);
      }
      sscanf((char*)sz_CurLine,"%X",(unsigned int*)&i_CurAddr);

   } while( (i_CurAddr <= ( projectSettings->sections[i_PLTIndex].addr + 
               projectSettings->sections[i_PLTIndex].size)) &&
         (i_CurPLT < MAX_PLT_ENTRIES) ) ;

#ifdef REV_DEBUG
   if(i_CurPLT >= MAX_PLT_ENTRIES)
   {
      DBG_PRINT("RE_GetPLTJumps: Warning, too much PLT entries.");
   }
#endif  
   fclose(f_AsmFile);
   DBG_PRINT("End of RE_GetPLTJumps...\n");  
   return(Z_OK);
}

/***********************************************************
 *  Function: RE_GetDynSymsRealAddr        File: reveng.c  *
 ***********************************************************
 * Usage:
 *
 * Put the real call address into dynamic symbols by
 * reading the plt section.
 ***********************************************************/ 
int32 RE_GetDynSymsRealAddr(LIST_Mapping *projectMapping, PLT_Jump *projectPLT)
{
   uint32 i;
   uint8 b_End;
   Mapping *p_CurMap;

   DBG_PRINT("Entering RE_GetDynSymsRealAddr...\n");  

   /* if no symbols then return */
   if(projectMapping[MAP_DYNSYM].list_Map == NULL)
      return(Z_OK);

   i=0;
   while(projectPLT[i].addr != 0)
   {
      b_End = Z_FALSE;
      p_CurMap = projectMapping[MAP_DYNSYM].list_Map;
      while(b_End == Z_FALSE)
      {
         if(p_CurMap->addr == projectPLT[i].value)
         {
            p_CurMap->addr = projectPLT[i].addr;
            b_End = Z_TRUE;
         }

         p_CurMap = p_CurMap->next;
         if(p_CurMap == NULL)
            b_End = Z_TRUE;
      }    
      if(p_CurMap == NULL)
         DBG_PRINT("Orphan Plt\n"); /*, projectPLT[i].value,
                     projectPLT[i].addr));*/
      i++;
   }

   DBG_PRINT("End of RE_GetPLTJumps...\n");
   return(Z_OK);
}

/***********************************************************
 *  Function: RE_GetMappingReferences      File: reveng.c  *
 ***********************************************************
 * Usage:
 *
 * Gets alls references of objects that are in the mapping.
 ***********************************************************/ 
int32 RE_GetMappingReferences(Settings *projectSettings, 
                              LIST_Mapping *projectMapping)
{
   FILE            *f_AsmFile, *f_ExeFile;  
   int32           i_ret,i;
   RE_ResultType   i_ResultType;
   Function_Params func_Params;
   uint8           sz_CurLine[STRING_MAX_SIZE];
   uint8           *psz_Inst;
   uint32          i_CurAddr, i_Result;
   uint32          i_PLTIndex, i_RodataIndex, i_InitIndex;
   uint32          i_AddrStart, i_AddrEnd;
   uint8           *pba_ReferenceExists;
   uint8           *pba_JmpReferenceExists;
   Mapping         *curMapping, *curFunction;
   Reference       curRef;
   Register        registers[REG_MAX];
   RegisterContext context;

   DBG_PRINT("Entering RE_GetMappingReferences...\n");  

   /* initialisation */
   curFunction = NULL;
   curRef.next = NULL;
   func_Params.isValid = Z_FALSE;
   RE_InitRegisters(registers, Z_TRUE); 

   RE_EmptyCntx(&context);

   /* searches for plt section */
   i_PLTIndex = MAX_SECTIONS;
   for(i=0; i< MAX_SECTIONS; i++)
   {
      if( strcmp((char*)(projectSettings->sections[i].name),".plt") == 0)
      {
         i_PLTIndex = i;
         break;
      }
   }

   /* searches for rodata section to search strings */
   i_RodataIndex = MAX_SECTIONS;
   for(i=0; i< MAX_SECTIONS; i++)
   {
      if( strcmp((char*)(projectSettings->sections[i].name),".rodata") == 0)
      {
         i_RodataIndex = i;
         break;
      }
   }
   /* if there is no rodata section, we will search in data section */
   if(i_RodataIndex == MAX_SECTIONS)
   {
      for(i=0; i< MAX_SECTIONS; i++)
      {
         if( strcmp((char*)(projectSettings->sections[i].name),".data") == 0)
         {
            i_RodataIndex = i;
            break;
         }
      }
   }

   /* searches for init section */
   i_InitIndex = MAX_SECTIONS;
   for(i=0; i< MAX_SECTIONS; i++)
   {
      if( strcmp((char*)(projectSettings->sections[i].name),".init") == 0)
      {
         i_InitIndex = i;
         break;
      }
   }

   /*creates an array of valid addresses to improve search speed */
   curMapping = projectMapping[MAP_SYMBOLS].list_Map;
   i_AddrStart = curMapping->addr; 
   if( (i_InitIndex != MAX_SECTIONS) &&
       (projectSettings->sections[i_InitIndex].addr < i_AddrStart) )
   {
      i_AddrStart = projectSettings->sections[i_InitIndex].addr;
   }
   while(curMapping->next != NULL)
   {
      curMapping = curMapping->next;
   }
   i_AddrEnd = curMapping->addr;

   pba_ReferenceExists = (uint8*) calloc(i_AddrEnd - i_AddrStart + 1, 1);
   pba_JmpReferenceExists = (uint8*) calloc(i_AddrEnd - i_AddrStart + 1, 1);  
   curMapping = projectMapping[MAP_SYMBOLS].list_Map;
   while(curMapping != NULL)
   {
      pba_ReferenceExists[curMapping->addr - i_AddrStart] = Z_TRUE;
      curMapping = curMapping->next;
   }  
   /* end initialisation */     

   f_AsmFile =  fopen((char*)(projectSettings->asm_SrcFileName),"r"); 
   if(f_AsmFile == NULL)
      return(FILE_ERROR_OPEN);
   
   f_ExeFile =  fopen((char*)(projectSettings->exe_FileName),"r");  
   if(f_ExeFile == NULL)
      return(FILE_ERROR_OPEN);

   /* get next function addr */
   curMapping = projectMapping[MAP_SYMBOLS].list_Map;
   while( (curMapping != NULL) &&
         (curMapping->type != OBJ_TYPE_FUNCTION) )
   {
      curMapping = curMapping->next;
   }

   i_ret = (uint32)fgets((char*)sz_CurLine,STRING_MAX_SIZE,f_AsmFile);
   /* process all asm lines */
   while(i_ret != 0)
   {
      sscanf((char*)sz_CurLine,"%X",(unsigned int*)&i_CurAddr);  
      /* don't process plt section */  
      if( (i_PLTIndex != MAX_SECTIONS) &&
          (i_CurAddr >= projectSettings->sections[i_PLTIndex].addr) &&
          (i_CurAddr < projectSettings->sections[i_PLTIndex].addr + 
           projectSettings->sections[i_PLTIndex].size) )
      {
         i_ret = (uint32)fgets((char*)sz_CurLine,STRING_MAX_SIZE,f_AsmFile);
         continue;
      }

      /* goto instruction label */
      psz_Inst = sz_CurLine;
      for(i=0; i<2; i++)
      {
         while( (*psz_Inst != ' ') &&
                (*psz_Inst != '\0') )
         {
            psz_Inst++;
         }

         while( (*psz_Inst == ' ') &&
                (*psz_Inst != '\0') )
         {
            psz_Inst++;
         }
         if(*psz_Inst == '\0')
         {
            i_ret = (uint32)fgets((char*)sz_CurLine,STRING_MAX_SIZE,f_AsmFile);
            continue;
         }
      }

      /* Check if we are at the begining of a new function. If new function then 
         get parameters and clean registers */
      if(curMapping != NULL)
      {
         if(curMapping->addr < i_CurAddr) /* we missed the start of the func */
         {            
            do
            {
               do
               {
                  curMapping = curMapping->next;
               } while( (curMapping != NULL) &&
                        (curMapping->type != OBJ_TYPE_FUNCTION) );
            } while( (curMapping != NULL) && (curMapping->addr < i_CurAddr) );
         }
      }
      
      if(curMapping != NULL)
      {
         if(curMapping->addr == i_CurAddr)
         {
            curFunction = curMapping;
            RE_InitRegisters(registers, Z_TRUE);
            RE_EmptyCntx(&context);
            RE_GetParameters(curMapping->label, &func_Params, 
                             &(projectMapping[MAP_PARAMS]));
            curMapping = curMapping->next;
            while( (curMapping != NULL) &&
                  (curMapping->type != OBJ_TYPE_FUNCTION) )
            {
               curMapping = curMapping->next;
            }
         }
      }

      /* analyse the instruction to spy registers and datas */
      ASM_ProcessInstruction(registers, &context, psz_Inst, curFunction,
                             i_CurAddr, &i_Result, &i_ResultType,
                             func_Params.functionHasMessage,
                             &func_Params);
      if(i_ResultType != RESULT_TYPE_INVALID)
      {
         /* ugly function with too much params! */
         RE_AnalyseResult(psz_Inst, 
                          i_Result, i_ResultType,
                          i_CurAddr, i_AddrStart, i_AddrEnd,
                          i_PLTIndex, i_RodataIndex,
                          projectSettings, 
                          projectMapping,
                          pba_ReferenceExists,
                          pba_JmpReferenceExists,
                          f_ExeFile,
                          &func_Params, curFunction);
      } /* if (i_ResultType != RESULT_TYPE_INVALID) */

      i_ret = (uint32)fgets((char*)sz_CurLine,STRING_MAX_SIZE,f_AsmFile);
   } /* while(i_ret != 0) */ 

   fclose(f_ExeFile);
   fclose(f_AsmFile);
   free(pba_ReferenceExists);
   free(pba_JmpReferenceExists);
   DBG_PRINT("End of RE_GetMappingReferences\n");    
   return(Z_OK);
}

/***********************************************************
 *  Function: RE_AnalyseResult             File: reveng.c  *
 ***********************************************************
 * Usage:
 *
 * Gets the reference for the result if one exists.
 ***********************************************************/ 
int32 RE_AnalyseResult(uint8 *sz_Instruction, 
      uint32 i_Result, RE_ResultType i_ResultType,
      uint32 i_CurAddr, uint32 i_AddrStart, uint32 i_AddrEnd,
      uint32 i_PLTIndex, uint32 i_TextSection,
      Settings *projectSettings, 
      LIST_Mapping *projectMapping,
      uint8 *pba_ReferenceExists,
      uint8 *pba_JmpReferenceExists,
      FILE *f_ExeFile,
      Function_Params *func_Params,
      Mapping *curFunction)
{
   int32 i_ret;
   uint32 sz_Message;  
   uint8 sz_String[STRING_MAX_SIZE];  
   uint8 sz_Buffer[STRING_MAX_SIZE];  
   uint8 b_Found;
   uint8 i_ParamIndex;
   Mapping newMapping, *curMapping, messageMapping;
   Reference curRef;  
   
   /* initialisation */
   b_Found = Z_FALSE;
   /* end initialisation */

   if(i_ResultType == RESULT_TYPE_MESSAGE)
   {
      i_ret = RE_GetMessage(&(projectMapping[MAP_MESSAGES]), 
                            i_Result, &messageMapping);
      if(i_ret == Z_OK)
      {
         sz_Message = INT2STR(i_Result); 
         sprintf((char*)(newMapping.label),"; Message \'%.4s\' (%s)", 
               (uint8*)&sz_Message, messageMapping.label);
      }
      else
      {
         sz_Message = INT2STR(i_Result); 
         sprintf((char*)(newMapping.label),"; Message \'%.4s\'", 
                 (uint8*)&sz_Message);    
      }

      /* adds reference in the mapping */
      newMapping.addr = i_CurAddr;
      newMapping.type = OBJ_TYPE_MESSAGE;
      newMapping.references.list_Ref = NULL;
      LIST_AddMap(&(projectMapping[MAP_REFS]), newMapping);
      return(Z_OK);
   }
   else if( (i_ResultType == RESULT_TYPE_PARAM) &&
            (func_Params->isValid == Z_TRUE) )
   {
      i_ParamIndex = RE_GetParamIndex(func_Params, (uint16)i_Result);
      if(i_ParamIndex != 0)
      {
         /* adds reference in the mapping */
         newMapping.addr = i_CurAddr;
         newMapping.type = OBJ_TYPE_VAR_REFERENCE;
         newMapping.references.list_Ref = NULL;
         if(func_Params->functionIsC == Z_TRUE)
            sprintf((char*)(newMapping.label),"; param %u (%s)", i_ParamIndex, 
                    func_Params->param[i_ParamIndex-1].name);
         else
            sprintf((char*)(newMapping.label),"; param %u (%s)", i_ParamIndex - 1, 
                    func_Params->param[i_ParamIndex-1].name);

         LIST_AddMap(&(projectMapping[MAP_REFS]), newMapping);
         return(Z_OK);
      }
   }
   else if( (STREQ4(sz_Instruction,"call")) ||
            (STREQ4(sz_Instruction,"push")) )
   {
      /* if i_Result is in the PLT section... */
      if( (i_PLTIndex != MAX_SECTIONS) &&
          (i_Result >= projectSettings->sections[i_PLTIndex].addr) &&
          (i_Result < projectSettings->sections[i_PLTIndex].addr + 
                      projectSettings->sections[i_PLTIndex].size) )
      {
         curMapping = projectMapping[MAP_DYNSYM].list_Map;
         while((curMapping != NULL) &&
               (b_Found == Z_FALSE) )
         {
            if(curMapping->addr > i_Result)
               break;

            /* found an object */
            else if(curMapping->addr == i_Result)
               b_Found = Z_TRUE;
            else
               curMapping = curMapping->next;
         } /* while((curMapping != NULL) &&... */      

         if(b_Found == Z_TRUE)
         {
            Mapping *p_NewMap;
            
            /* adds reference in the mapping */
            newMapping.addr = i_CurAddr;
            newMapping.type = OBJ_TYPE_REFERENCE;
            newMapping.references.list_Ref = NULL;
            sprintf((char*)sz_Buffer,"function \"%s\"",curMapping->label);

            if(i_ResultType == RESULT_TYPE_DIRECT_VALUE)
            {
               sprintf((char*)(newMapping.label),"%sReference to %s",
                        projectSettings->blank_line, sz_Buffer);
            }
            else
            {
               sprintf((char*)(newMapping.label),"%sPossible reference to %s",
                       projectSettings->blank_line, sz_Buffer);
            }
            p_NewMap = LIST_AddMap(&(projectMapping[MAP_REFS]), newMapping);
            if( (p_NewMap != NULL) && (projectSettings->printGraph == Z_TRUE) )
            {
               if(curFunction != NULL)
                  LIST_AddCallee(&(curFunction->callees), curMapping);
            }
            return(Z_OK);
         }
      } /* if i_Result... */
   } /* end if call */

   /* searches for a match in the mapping */     
   if( (i_Result >= i_AddrStart) && (i_Result <= i_AddrEnd) &&
         (pba_ReferenceExists[i_Result - i_AddrStart] == Z_TRUE) )
   {
      curMapping = projectMapping[MAP_SYMBOLS].list_Map;
      while((curMapping != NULL) &&
            (b_Found == Z_FALSE) )
      {
         if(curMapping->addr > i_Result)
            break;

         /* found an object */
         else if( (curMapping->addr == i_Result) &&
               (curMapping->type != OBJ_TYPE_REFERENCE) )
         {
            curRef.addr = i_CurAddr;              
            if(i_ResultType == RESULT_TYPE_DIRECT_VALUE)
               curRef.type = DIRECT_REF;
            else
               curRef.type = INDIRECT_REF;

            LIST_AppendRef(&(curMapping->references), curRef);                       
            b_Found = Z_TRUE;
         }
         else
         {
            curMapping = curMapping->next;
         }
      } /* while((curMapping != NULL) &&... */      

      if(b_Found == Z_TRUE)
      {
         Mapping *p_NewMap;

         if(curMapping->type == OBJ_TYPE_INSTRUCTION)
            return(Z_OK);

         /* adds reference in the mapping */
         newMapping.addr = i_CurAddr;
         newMapping.type = OBJ_TYPE_REFERENCE;
         newMapping.references.list_Ref = NULL;
         if(curMapping->type == OBJ_TYPE_OBJECT)
            sprintf((char*)sz_Buffer,"object \"%s\"",curMapping->label);
         else if(curMapping->type == OBJ_TYPE_FUNCTION)
            sprintf((char*)sz_Buffer,"function \"%s\"",curMapping->label);

         if(i_ResultType == RESULT_TYPE_DIRECT_VALUE)
         {
            sprintf((char*)(newMapping.label),"%sReference to %s",
                     projectSettings->blank_line, sz_Buffer);
         }
         else
         {
            sprintf((char*)(newMapping.label),"%sPossible reference to %s",
                     projectSettings->blank_line, sz_Buffer);
         }

         p_NewMap = LIST_AddMap(&(projectMapping[MAP_REFS]), newMapping);
         if( (p_NewMap != NULL) && 
             (projectSettings->printGraph == Z_TRUE) &&
             (curMapping->type == OBJ_TYPE_FUNCTION))
         {
            if(curFunction != NULL)
               LIST_AddCallee(&(curFunction->callees), curMapping);
         }
      }
   } /* (pba_ReferenceExists[i_Result - i_AddrStart] == Z_TRUE) */
   /* if jump and not found */
   else if( (i_ResultType == RESULT_TYPE_JUMP) &&
         (i_Result >= i_AddrStart) && (i_Result <= i_AddrEnd) )
   {
      /* gets mapping ptr */    
      if(pba_JmpReferenceExists[i_Result - i_AddrStart] == Z_FALSE)
      {
         sprintf((char*)(newMapping.label),"\ninstruction");
         newMapping.addr = i_Result;
         newMapping.type = OBJ_TYPE_INSTRUCTION;
         newMapping.references.list_Ref = NULL;
         LIST_AddMap(&(projectMapping[MAP_JUMPS]), newMapping);
         pba_JmpReferenceExists[i_Result - i_AddrStart] = Z_TRUE;
      }

      curMapping = projectMapping[MAP_JUMPS].tail;
      while((curMapping != NULL) &&
            (curMapping->addr != i_Result) )
      {
         curMapping = curMapping->prev;
      }

      /* adds the current reference */
      curRef.addr = i_CurAddr;              
      if(i_ResultType == RESULT_TYPE_DIRECT_VALUE)
         curRef.type = DIRECT_REF;
      else
         curRef.type = INDIRECT_REF;

      LIST_AppendRef(&(curMapping->references), curRef);
      return(Z_OK);
   }
   
   /* Always Search for a string in text section because a reference to an
      object may also be a pointer to a string */
   if( (i_TextSection != MAX_SECTIONS) &&
       (i_Result >= projectSettings->sections[i_TextSection].addr) &&
       (i_Result < projectSettings->sections[i_TextSection].addr + 
          projectSettings->sections[i_TextSection].size) )
   {
      i_ret = FILE_GetStringAt(f_ExeFile, i_Result, sz_String);
      /* if the result is a valid string then we read it 
         and add it to the mapping. */
      if( (i_ret == Z_OK) && 
          (sz_String[0] != '\0') &&
          (sz_String[0] != '\n') )
      {
         /* looks for the string in the mapping */
         curMapping = projectMapping[MAP_STRINGS].list_Map;
         while(curMapping != NULL)
         {
            if(curMapping->addr == i_Result)
               break;
            curMapping = curMapping->next;
         }
         /* if string is not in the list, add it */
         if(curMapping == NULL)
         {
            if(sz_String[strlen((char*)sz_String)-1] == '\n')
               sz_String[strlen((char*)sz_String)-1] = '\0';

            strcpy((char*)(newMapping.label), (char*)sz_String);
            newMapping.addr = i_Result;
            newMapping.type = OBJ_TYPE_STRING;
            LIST_AddMap(&projectMapping[MAP_STRINGS], newMapping);
         }

         /* adds a new reference to the string */
         if(i_ResultType == RESULT_TYPE_DIRECT_VALUE)
         {
            if(b_Found == Z_FALSE)
            {
               sprintf((char*)(newMapping.label),"%sReference to string \"%s\"",
                     projectSettings->blank_line, sz_String);
            }
            else  /* a reference on this adress already exists: no blank lite before */
            {
               sprintf((char*)(newMapping.label),"Reference to string \"%s\"", sz_String);
            }
         }
         else
         {
            if(b_Found == Z_FALSE)
            {
               sprintf((char*)(newMapping.label),"%sPossible reference to string \"%s\"",
                     projectSettings->blank_line, sz_String);
            }
            else  /* a reference on this adress already exists: no blank lite before */
            {
               sprintf((char*)(newMapping.label),"Possible reference to string \"%s\"", 
                        sz_String);
            }
         }

         newMapping.addr = i_CurAddr;
         newMapping.type = OBJ_TYPE_REFERENCE;
         newMapping.references.list_Ref = NULL;
         LIST_AddMap(&(projectMapping[MAP_REFS]), newMapping);
      }
      return(Z_OK);
   }       

   return(Z_OK);
}

/***********************************************************
 *  Function: RE_GetMessage                File: reveng.c  *
 ***********************************************************
 * Usage:
 *
 * Gets the message structure if i_Message exists.
 ***********************************************************/ 
int32 RE_GetMessage(LIST_Mapping *messagesMapping, uint32 i_Message, Mapping *resultMapping)
{
   Mapping *p_CurMapping;

   p_CurMapping = messagesMapping->list_Map;
   while(p_CurMapping != NULL)
   {
      if(p_CurMapping->addr == i_Message)
      {
         memcpy(resultMapping, p_CurMapping, sizeof(Mapping));
         return(Z_OK);
      }

      if(p_CurMapping->addr > i_Message)
         return(CHAIN_ERROR_NOT_FOUND);

      p_CurMapping = p_CurMapping->next;
   }

   return(CHAIN_ERROR_NOT_FOUND);
}
